import re
import math
'''
Identifier tools
'''


def id2float(name):
    m = re.search(r'[a-zA-Z]*(\d+\.?\d*)', name)
    if m:
        return float(m.group(1))
    else:
        print(r"[GREW]Node identifier '%s' does not follow grammar [a-zA-Z]*(\d+\.?\d*)" % name)


def float2id(f):
    a = int(f) if f.is_integer() else f
    return 'W%s' % a


def id_lt(n1, n2):
    return id2float(n1) < id2float(n2)


def mid(n1, n2):
    return float2id((id2float(n1) + id2float(n2)/2))


def left(n1):
    f1 = id2float(n1)
    if f1.is_integer():
        return float2id(f1-1)
    else:
        return float2id(math.floor(f1))


def right(n1):
    f1 = id2float(n1)
    if f1.is_integer():
        return float2id(f1+1)
    else:
        return float2id(math.ceil(f1))



''' Graph utility tools '''

class GrewError(Exception):
    """A wrapper for grew-related errors"""

    def __init__(self, message):
        self.value = message
    def __str__(self):
        return repr(self.value)

GrewError.__doc__ = "A wrapper for grew-related errors"


def glb(gr, pivot):
    """
    Greatest lower bound: the greatest node lower than [pivot] in [gr]
    :param gr: the dict-graph
    :param pivot: the pivot node
    :return: the greatest lower bound if it exists, otherwise None
    """
    res = None
    for nid in gr:
        if res and id_lt(res, nid) and id_lt(nid, pivot):
            res = nid
        elif id_lt(nid, pivot):
            res = nid
    return res


def lub(gr, pivot):
    """
    Least upper bound: the smallest node greater than [pivot] in [gr]
    :param gr: the dict-graph
    :param pivot: the pivot node
    :return: the least upper bound if it exists, otherwise None
    """
    res = None
    for nid in gr:
        if res and id_lt(nid, res) and id_lt(pivot, nid):
            res = nid
        elif id_lt(pivot, nid):
            res = nid
    return res



"""
Grew module : anything you want to talk about graphs
Graphs are represented either by a dict (called dict-graph),
or by an str (str-graph).
"""
import os.path
import re
import copy
import tempfile
import json

import utils
import network
import ast

''' Library tools '''

def init():
    """
    Initialize connection to GREW library
    :return: True if anything went right
    """
    return network.init()

def grs(grsystem):
    """Load a grs stored in a file
    :param grs: either a file name or a GREW string representation of a grs
    :raise an error if the file was not correctly loaded
    """
    if os.path.isfile(grsystem):
        req = { "command": "load_grs", "filename": grsystem }
        reply = network.send_and_receive(req)
    else:
        with tempfile.NamedTemporaryFile(mode="w", delete=True) as f:
            f.write(grsystem)
            f.seek(0)  # to be read by others
            req = { "command": "load_grs", "filename": fname }
            reply = network.send_and_receive(req)
    return reply["index"]

def run(rs, gr, strat="main"):
    """
    Apply rs or the last loaded one to [gr]
    :param gr: the graph, either a str (in grew format) or a dict
    :param rs: a graph rewriting
     system or a GREW string representation of a grs
    :param strategy: the strategy. By default, !
    :return: the list of rewritten graphs
    """
    req = {
        "command": "run",
        "graph": json.dumps(gr),
        "grs_index": rs,
        "strat": strat
    }
    reply = network.send_and_receive(req)
    return reply

def search(pattern, gr):
    """
    Search for [pattern] into [gr]
    :param patten: a string pattern
    :param gr: the graph
    :return: the list of matching of [pattern] into [gr]
    """
    req = {
        "command": "search",
        "graph": json.dumps(gr),
        "pattern": pattern
    }
    reply = network.send_and_receive(req)
    return reply

def json_grs(rs):
    req = { "command": "json_grs", "grs_index": rs }
    return network.send_and_receive(req)

def dot_graph(gr):
    req = { "command": "dot_graph", "graph": json.dumps(gr) }
    return network.send_and_receive(req)




from distutils.core import setup

setup(
    name='Grew',
    version='0.1.1',
    packages=['grew',],
    license='LICENSE/Licence_CeCILL_V2-en.txt',
    long_description=open('README').read(),
    url="http://pypi.python.org/pypi/Grew",
    author="bguil",
    author_email="Bruno.Guillaume@loria.fr"
)



''' Utility tools to connect to ocaml GREW'''

import subprocess
import time
import socket
import os.path
import json
import utils
import grew

host = 'localhost'
port = 8888
remote_ip = ''

def init():
    global port, remote_ip
    python_pid = os.getpid()
    while (port<8898):
        caml = subprocess.Popen(["grewpy_daemon", "--caller", str(python_pid), "--port", str(port)])
        #wait for grew's lib answer
        time.sleep(0.1)
        if caml.poll() == None:
            print ("connected to port: " + str(port))
            remote_ip = socket.gethostbyname(host)
            return (caml)
        else:
            port += 1
    print ("Failed to connect 10 times!")
    exit (1)

def connect():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((remote_ip, port))
        return s
    except socket.error:
        print('[GREW] Failed to create socket. Sorry\n')
    except socket.gaierror:
        print('[GREW] Hostname could not be resolved. Sorry\n')


def send_and_receive(msg):
    try:
        stocaml = connect()
        json_msg = json.dumps(msg)
        stocaml.sendall(json_msg.encode(encoding='UTF-8'))
        camltos = stocaml.recv(4096)
        stocaml.close()
        reply = json.loads(camltos.decode(encoding='UTF-8'))
        if reply["status"] == "OK":
            return reply["data"]
        elif reply["status"] == "ERROR":
            raise utils.GrewError (reply["message"])
    except socket.error:
        raise grew.GrewError('Library call socket error')
    except AttributeError: # connect issue
        raise grew.GrewError('Library call AttributeError')




import grew
import network

def _compute_next(rs,todo):
    next = set()
    for g in todo:
        sucs_g = grew.run(rs,g)
        if sucs_g != [g]:
            next |= set(sucs_g)
    return next

#latex is_terminating
def is_terminating(rs,gr):
    def loop(todo, seen):
        next = _compute_next(rs, todo)
        return next == set() or (not next.intersection(seen) and
                                    loop(next, seen.union(next)))
    return loop(set([gr]), set([gr]))
#/latex

